/*
 * System call numbers.
 *
 * DO NOT EDIT-- this file is automatically generated.
 * created from FreeBSD: src/sys/i386/ibcs2/syscalls.isc,v 1.5 1999/08/28 00:44:01 peter Exp 
 */

#define	IBCS2_ISC_ibcs2_rename	2
#define	IBCS2_ISC_ibcs2_sigaction	3
#define	IBCS2_ISC_ibcs2_sigprocmask	4
#define	IBCS2_ISC_ibcs2_sigpending	5
#define	IBCS2_ISC_getgroups	6
#define	IBCS2_ISC_setgroups	7
#define	IBCS2_ISC_ibcs2_pathconf	8
#define	IBCS2_ISC_ibcs2_fpathconf	9
#define	IBCS2_ISC_ibcs2_wait	11
#define	IBCS2_ISC_setsid	12
#define	IBCS2_ISC_getpid	13
#define	IBCS2_ISC_ibcs2_sysconf	16
#define	IBCS2_ISC_ibcs2_sigsuspend	17
#define	IBCS2_ISC_ibcs2_symlink	18
#define	IBCS2_ISC_ibcs2_readlink	19
#define	IBCS2_ISC_MAXSYSCALL	21
