/* $FreeBSD: src/sys/alpha/tc/tcdevs_data.h,v 1.2 1999/08/28 00:39:11 peter Exp $ */
/*	$NetBSD: tcdevs_data.h,v 1.12 1998/06/18 23:29:46 jonathan Exp $	*/

/*
 * THIS FILE AUTOMATICALLY GENERATED.  DO NOT EDIT.
 *
 * generated from:
 *	NetBSD: tcdevs,v 1.10 1998/06/18 23:28:12 jonathan Exp 
 */

/*
 * Copyright (c) 1996 Christopher G. Demetriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Christopher G. Demetriou
 *	for the NetBSD Project.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

struct tc_knowndev tc_knowndevs[] = {
	{
	    "AV01B-AA",
	    TC_DEVICE_AV01B_AA,
	    TC_DESCRIPTION_AV01B_AA,
	},
	{
	    "DGLTA-FA",
	    TC_DEVICE_DGLTA_FA,
	    TC_DESCRIPTION_DGLTA_FA,
	},
	{
	    "FORE_ATM",
	    TC_DEVICE_FORE_ATM,
	    TC_DESCRIPTION_FORE_ATM,
	},
	{
	    "KZTSA-AA",
	    TC_DEVICE_KZTSA_AA,
	    TC_DESCRIPTION_KZTSA_AA,
	},
	{
	    "OTTO    ",
	    TC_DEVICE_OTTO,
	    TC_DESCRIPTION_OTTO,
	},
	{
	    "PMAD-AA ",
	    TC_DEVICE_PMAD_AA,
	    TC_DESCRIPTION_PMAD_AA,
	},
	{
	    "PMAF-AA ",
	    TC_DEVICE_PMAF_AA,
	    TC_DESCRIPTION_PMAF_AA,
	},
	{
	    "PMAF-F  ",
	    TC_DEVICE_PMAF_F,
	    TC_DESCRIPTION_PMAF_F,
	},
	{
	    "PMAG-AA ",
	    TC_DEVICE_PMAG_AA,
	    TC_DESCRIPTION_PMAG_AA,
	},
	{
	    "PMAG-BA ",
	    TC_DEVICE_PMAG_BA,
	    TC_DESCRIPTION_PMAG_BA,
	},
	{
	    "PMAG-CA ",
	    TC_DEVICE_PMAG_CA,
	    TC_DESCRIPTION_PMAG_CA,
	},
	{
	    "PMAG-DA ",
	    TC_DEVICE_PMAG_DA,
	    TC_DESCRIPTION_PMAG_DA,
	},
	{
	    "PMAG-DV ",
	    TC_DEVICE_PMAG_DV,
	    TC_DESCRIPTION_PMAG_DV,
	},
	{
	    "PMAG-FA ",
	    TC_DEVICE_PMAG_FA,
	    TC_DESCRIPTION_PMAG_FA,
	},
	{
	    "PMAGB-BA",
	    TC_DEVICE_PMAGB_BA,
	    TC_DESCRIPTION_PMAGB_BA,
	},
	{
	    "PMAGD   ",
	    TC_DEVICE_PMAGD,
	    TC_DESCRIPTION_PMAGD,
	},
	{
	    "PMAGD-AA",
	    TC_DEVICE_PMAGD_AA,
	    TC_DESCRIPTION_PMAGD_AA,
	},
	{
	    "PMAGD-BA",
	    TC_DEVICE_PMAGD_BA,
	    TC_DESCRIPTION_PMAGD_BA,
	},
	{
	    "PMAZ-AA ",
	    TC_DEVICE_PMAZ_AA,
	    TC_DESCRIPTION_PMAZ_AA,
	},
	{
	    "PMAZ-DS ",
	    TC_DEVICE_PMAZ_DS,
	    TC_DESCRIPTION_PMAZ_DS,
	},
	{
	    "PMAZ-FS ",
	    TC_DEVICE_PMAZ_FS,
	    TC_DESCRIPTION_PMAZ_FS,
	},
	{
	    "PMAZB-AA",
	    TC_DEVICE_PMAZB_AA,
	    TC_DESCRIPTION_PMAZB_AA,
	},
	{
	    "PMAZC-AA",
	    TC_DEVICE_PMAZC_AA,
	    TC_DESCRIPTION_PMAZC_AA,
	},
	{
	    "T1D4PKT ",
	    TC_DEVICE_T1D4PKT,
	    TC_DESCRIPTION_T1D4PKT,
	},
	{
	    "T3PKT   ",
	    TC_DEVICE_T3PKT,
	    TC_DESCRIPTION_T3PKT,
	},
	{
	    "PMAT-AA ",
	    TC_DEVICE_PMAT_AA,
	    TC_DESCRIPTION_PMAT_AA,
	},
	{ NULL, NULL, NULL, }
};
