/*
 * System call numbers.
 *
 * DO NOT EDIT-- this file is automatically generated.
 * created from FreeBSD: src/sys/i386/ibcs2/syscalls.xenix,v 1.6 1999/08/28 00:44:02 peter Exp 
 */

#define	IBCS2_XENIX_xenix_rdchk	7
#define	IBCS2_XENIX_xenix_chsize	10
#define	IBCS2_XENIX_xenix_ftime	11
#define	IBCS2_XENIX_xenix_nap	12
#define	IBCS2_XENIX_xenix_scoinfo	21
#define	IBCS2_XENIX_select	36
#define	IBCS2_XENIX_xenix_eaccess	37
#define	IBCS2_XENIX_ibcs2_sigaction	39
#define	IBCS2_XENIX_ibcs2_sigprocmask	40
#define	IBCS2_XENIX_ibcs2_sigpending	41
#define	IBCS2_XENIX_ibcs2_sigsuspend	42
#define	IBCS2_XENIX_ibcs2_getgroups	43
#define	IBCS2_XENIX_ibcs2_setgroups	44
#define	IBCS2_XENIX_ibcs2_sysconf	45
#define	IBCS2_XENIX_ibcs2_pathconf	46
#define	IBCS2_XENIX_ibcs2_fpathconf	47
#define	IBCS2_XENIX_ibcs2_rename	48
#define	IBCS2_XENIX_xenix_utsname	50
#define	IBCS2_XENIX_getitimer	55
#define	IBCS2_XENIX_setitimer	56
#define	IBCS2_XENIX_MAXSYSCALL	57
