/* misc IOCTL's */

/* Unix to Jerq I/O control */
#define IBCS2_JTYPE             ('j'<<8)
#define IBCS2_JBOOT             (IBCS2_JTYPE|1)
#define IBCS2_JTERM             (IBCS2_JTYPE|2)
#define IBCS2_JMPX              (IBCS2_JTYPE|3)
#define IBCS2_JTIMO             (IBCS2_JTYPE|4)
#define IBCS2_JWINSIZE          (IBCS2_JTYPE|5)
#define IBCS2_JTIMOM            (IBCS2_JTYPE|6)
#define IBCS2_JZOMBOOT          (IBCS2_JTYPE|7)
#define IBCS2_JAGENT            (IBCS2_JTYPE|9)
#define IBCS2_JTRUN             (IBCS2_JTYPE|10)


/* Defines for keyboard and display ioctl's */
#define IBCS2_KIOC              ('K'<<8)
#define IBCS2_KDDISPTYPE        (IBCS2_KIOC|1)
#define IBCS2_KDMAPDISP         (IBCS2_KIOC|2)
#define IBCS2_KDUNMAPDISP       (IBCS2_KIOC|3)
#define IBCS2_KDGKBENT          (IBCS2_KIOC|4)
#define IBCS2_KDSKBENT          (IBCS2_KIOC|5)
#define IBCS2_KDGKBMODE         (IBCS2_KIOC|6)
#define IBCS2_KDSKBMODE         (IBCS2_KIOC|7)
#define IBCS2_KDMKTONE          (IBCS2_KIOC|8)
#define IBCS2_KDGETMODE         (IBCS2_KIOC|9)
#define IBCS2_KDSETMODE         (IBCS2_KIOC|10)
#define IBCS2_KDADDIO           (IBCS2_KIOC|11)
#define IBCS2_KDDELIO           (IBCS2_KIOC|12)
#define IBCS2_KDSBORDER         (IBCS2_KIOC|13)
#define IBCS2_KDQUEMODE         (IBCS2_KIOC|15)
#define IBCS2_KDVDCTYPE         (IBCS2_KIOC|16)
#define IBCS2_KDGKBSTATE        (IBCS2_KIOC|19)
#define IBCS2_KDSETRAD          (IBCS2_KIOC|20)
#define IBCS2_KDENABIO          (IBCS2_KIOC|60)
#define IBCS2_KDDISABIO         (IBCS2_KIOC|61)
#define IBCS2_KIOCINFO          (IBCS2_KIOC|62)
#define IBCS2_KIOCSOUND         (IBCS2_KIOC|63)
#define IBCS2_KDGKBTYPE         (IBCS2_KIOC|64)
#define IBCS2_KDGETLED          (IBCS2_KIOC|65)
#define IBCS2_KDSETLED          (IBCS2_KIOC|66)


/* Defines for Xenix keyboard and display ioctl's */
#define IBCS2_MIOC              ('k' << 8)
#define IBCS2_GETFKEY           (IBCS2_MIOC | 0)
#define IBCS2_SETFKEY           (IBCS2_MIOC | 1)
#define IBCS2_GIO_SCRNMAP       (IBCS2_MIOC | 2)
#define IBCS2_PIO_SCRNMAP       (IBCS2_MIOC | 3)
#define IBCS2_GIO_STRMAP_21     (IBCS2_MIOC | 4)
#define IBCS2_PIO_STRMAP_21     (IBCS2_MIOC | 5)
#define IBCS2_GIO_KEYMAP        (IBCS2_MIOC | 6)
#define IBCS2_PIO_KEYMAP        (IBCS2_MIOC | 7)
#define IBCS2_GIO_STRMAP        (IBCS2_MIOC | 11)
#define IBCS2_PIO_STRMAP        (IBCS2_MIOC | 12)
