/*	$FreeBSD: src/sys/dev/pccard/pccarddevs_data.h,v 1.3 1999/11/29 20:54:14 imp Exp $	*/

/*
 * THIS FILE AUTOMATICALLY GENERATED.  DO NOT EDIT.
 *
 * generated from:
 *	FreeBSD: src/sys/dev/pccard/pccarddevs,v 1.3 1999/11/29 20:53:36 imp Exp 
 */
/* $NetBSD: pcmciadevs,v 1.60 1999/11/29 02:28:19 jun Exp $ */

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Christos Zoulas.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

struct pccard_knowndev pccard_knowndevs[] = {
	{
	    PCCARD_VENDOR_ADAPTEC, PCCARD_PRODUCT_ADAPTEC_APA1460_1,
	    PCCARD_CIS_ADAPTEC_APA1460_1,
	    0,
	    "Adaptec Corporation",
	    "Adaptec APA-1460/A SCSI Host Adapter"	},
	},
	{
	    PCCARD_VENDOR_ADAPTEC, PCCARD_PRODUCT_ADAPTEC_APA1460_2,
	    PCCARD_CIS_ADAPTEC_APA1460_2,
	    0,
	    "Adaptec Corporation",
	    "Adaptec APA-1460/B SCSI Host Adapter"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3C562,
	    PCCARD_CIS_3COM_3C562,
	    0,
	    "3Com Corporation",
	    "3Com 3c562 33.6 Modem/10Mbps Ethernet"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3C589,
	    PCCARD_CIS_3COM_3C589,
	    0,
	    "3Com Corporation",
	    "3Com 3c589 10Mbps Ethernet"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3C574,
	    PCCARD_CIS_3COM_3C574,
	    0,
	    "3Com Corporation",
	    "3Com 3c574-TX 10/100Mbps Ethernet"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3CXM056BNW,
	    PCCARD_CIS_3COM_3CXM056BNW,
	    0,
	    "3Com Corporation",
	    "3Com/NoteWorthy 3CXM056-BNW 56K Modem"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3CXEM556,
	    PCCARD_CIS_3COM_3CXEM556,
	    0,
	    "3Com Corporation",
	    "3Com/Megahertz 3CXEM556 Ethernet/Modem"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3CXEM556INT,
	    PCCARD_CIS_3COM_3CXEM556INT,
	    0,
	    "3Com Corporation",
	    "3Com/Megahertz 3CXEM556-INT Ethernet/Modem"	},
	},
	{
	    PCCARD_VENDOR_3COM, PCCARD_PRODUCT_3COM_3CCFEM556BI,
	    PCCARD_CIS_3COM_3CCFEM556BI,
	    0,
	    "3Com Corporation",
	    "3Com/Megahertz 3CCFEM556BI Ethernet/Modem"	},
	},
	{
	    PCCARD_VENDOR_COMPEX, PCCARD_PRODUCT_COMPEX_LINKPORT_ENET_B,
	    PCCARD_CIS_COMPEX_LINKPORT_ENET_B,
	    0,
	    "Compex Corporation",
	    "Compex Linkport ENET-B Ethernet"	},
	},
	{
	    PCCARD_VENDOR_LASAT, PCCARD_PRODUCT_LASAT_CREDIT_288,
	    PCCARD_CIS_LASAT_CREDIT_288,
	    0,
	    "",
	    "Lasat Credit 288 Modem"	},
	},
	{
	    PCCARD_VENDOR_DAYNA, PCCARD_PRODUCT_DAYNA_COMMUNICARD_E_1,
	    PCCARD_CIS_DAYNA_COMMUNICARD_E_1,
	    0,
	    "Dayna Corporation",
	    "Dayna CommuniCard E"	},
	},
	{
	    PCCARD_VENDOR_DAYNA, PCCARD_PRODUCT_DAYNA_COMMUNICARD_E_2,
	    PCCARD_CIS_DAYNA_COMMUNICARD_E_2,
	    0,
	    "Dayna Corporation",
	    "Dayna CommuniCard E"	},
	},
	{
	    PCCARD_VENDOR_DIGITAL, PCCARD_PRODUCT_DIGITAL_MOBILE_MEDIA_CDROM,
	    PCCARD_CIS_DIGITAL_MOBILE_MEDIA_CDROM,
	    0,
	    "Digital Equipment Corporation",
	    "Digital Mobile Media CD-ROM"	},
	},
	{
	    PCCARD_VENDOR_FUJITSU, PCCARD_PRODUCT_FUJITSU_SCSI600,
	    PCCARD_CIS_FUJITSU_SCSI600,
	    0,
	    "Fujitsu Corporation",
	    "Fujitsu SCSI 600 Interface"	},
	},
	{
	    PCCARD_VENDOR_MOTOROLA, PCCARD_PRODUCT_MOTOROLA_POWER144,
	    PCCARD_CIS_MOTOROLA_POWER144,
	    0,
	    "Motorola Corporation",
	    "Motorola Power 14.4 Modem"	},
	},
	{
	    PCCARD_VENDOR_MOTOROLA, PCCARD_PRODUCT_MOTOROLA_PM100C,
	    PCCARD_CIS_MOTOROLA_PM100C,
	    0,
	    "Motorola Corporation",
	    "Motorola Personal Messenger 100C CDPD Modem"	},
	},
	{
	    PCCARD_VENDOR_FUJITSU, PCCARD_PRODUCT_FUJITSU_LA501,
	    PCCARD_CIS_FUJITSU_LA501,
	    0,
	    "Fujitsu Corporation",
	    "Fujitsu Towa LA501 Ethernet"	},
	},
	{
	    PCCARD_VENDOR_FUJITSU, PCCARD_PRODUCT_FUJITSU_LA10S,
	    PCCARD_CIS_FUJITSU_LA10S,
	    0,
	    "Fujitsu Corporation",
	    "Fujitsu Compact Flash Ethernet"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_3270,
	    PCCARD_CIS_IBM_3270,
	    0,
	    "IBM Corporation",
	    "3270 Emulation"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_INFOMOVER,
	    PCCARD_CIS_IBM_INFOMOVER,
	    0,
	    "IBM Corporation",
	    "National Semiconductor InfoMover"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_5250,
	    PCCARD_CIS_IBM_5250,
	    0,
	    "IBM Corporation",
	    "5250 Emulation"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_TROPIC,
	    PCCARD_CIS_IBM_TROPIC,
	    0,
	    "IBM Corporation",
	    "Token Ring 4/16"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_HOME_AND_AWAY,
	    PCCARD_CIS_IBM_HOME_AND_AWAY,
	    0,
	    "IBM Corporation",
	    "IBM Home and Away Modem"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_WIRELESS_LAN_ENTRY,
	    PCCARD_CIS_IBM_WIRELESS_LAN_ENTRY,
	    0,
	    "IBM Corporation",
	    "Wireless LAN Entry"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_PORTABLE_CDROM,
	    PCCARD_CIS_IBM_PORTABLE_CDROM,
	    0,
	    "IBM Corporation",
	    "PCMCIA Portable CD-ROM Drive"	},
	},
	{
	    PCCARD_VENDOR_IBM, PCCARD_PRODUCT_IBM_ETHERJET,
	    PCCARD_CIS_IBM_ETHERJET,
	    0,
	    "IBM Corporation",
	    "IBM EtherJet Ethernet"	},
	},
	{
	    PCCARD_VENDOR_IODATA, PCCARD_PRODUCT_IODATA_PCLATE,
	    PCCARD_CIS_IODATA_PCLATE,
	    0,
	    "I-O DATA",
	    "I-O DATA PCLA/TE"	},
	},
	{
	    PCCARD_VENDOR_LINKSYS, PCCARD_PRODUCT_LINKSYS_ECARD_1,
	    PCCARD_CIS_LINKSYS_ECARD_1,
	    0,
	    "Linksys Corporation",
	    "Linksys EthernetCard or D-Link DE-650"	},
	},
	{
	    PCCARD_VENDOR_LINKSYS, PCCARD_PRODUCT_LINKSYS_COMBO_ECARD,
	    PCCARD_CIS_LINKSYS_COMBO_ECARD,
	    0,
	    "Linksys Corporation",
	    "Linksys Combo EthernetCard"	},
	},
	{
	    PCCARD_VENDOR_LINKSYS, PCCARD_PRODUCT_LINKSYS_TRUST_COMBO_ECARD,
	    PCCARD_CIS_LINKSYS_TRUST_COMBO_ECARD,
	    0,
	    "Linksys Corporation",
	    "Trust (Linksys) Combo EthernetCard"	},
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ, PCCARD_PRODUCT_MEGAHERTZ_XJ4288,
	    PCCARD_CIS_MEGAHERTZ_XJ4288,
	    0,
	    "Megahertz Corporation",
	    "Megahertz XJ4288 Modem"	},
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ, PCCARD_PRODUCT_MEGAHERTZ_XJ4336,
	    PCCARD_CIS_MEGAHERTZ_XJ4336,
	    0,
	    "Megahertz Corporation",
	    "Megahertz XJ4336 Modem"	},
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ, PCCARD_PRODUCT_MEGAHERTZ_XJ5560,
	    PCCARD_CIS_MEGAHERTZ_XJ5560,
	    0,
	    "Megahertz Corporation",
	    "Megahertz X-JACK 56kbps Modem"	},
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ2, PCCARD_PRODUCT_MEGAHERTZ2_XJACK,
	    PCCARD_CIS_MEGAHERTZ2_XJACK,
	    0,
	    "Megahertz Corporation",
	    "Megahertz X-JACK Ethernet"	},
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ, PCCARD_PRODUCT_MEGAHERTZ_XJEM3336,
	    PCCARD_CIS_MEGAHERTZ_XJEM3336,
	    0,
	    "Megahertz Corporation",
	    "Megahertz X-JACK Ethernet Modem"	},
	},
	{
	    PCCARD_VENDOR_USROBOTICS, PCCARD_PRODUCT_USROBOTICS_WORLDPORT144,
	    PCCARD_CIS_USROBOTICS_WORLDPORT144,
	    0,
	    "US Robotics Corporation",
	    "US Robotics WorldPort 14.4 Modem"	},
	},
	{
	    PCCARD_VENDOR_SANDISK, PCCARD_PRODUCT_SANDISK_SDCFB,
	    PCCARD_CIS_SANDISK_SDCFB,
	    0,
	    "Sandisk Corporation",
	    "Sandisk CompactFlash Card"	},
	},
	{
	    PCCARD_VENDOR_SIMPLETECH, PCCARD_PRODUCT_SIMPLETECH_COMMUNICATOR288,
	    PCCARD_CIS_SIMPLETECH_COMMUNICATOR288,
	    0,
	    "Simple Technology",
	    "Simple Technology 28.8 Communicator"	},
	},
	{
	    PCCARD_VENDOR_SIMPLETECH, PCCARD_PRODUCT_SIMPLETECH_SPECTRUM24,
	    PCCARD_CIS_SIMPLETECH_SPECTRUM24,
	    0,
	    "Simple Technology",
	    "Symbol Spectrum24 WLAN Adapter"	},
	},
	{
	    PCCARD_VENDOR_SOCKET, PCCARD_PRODUCT_SOCKET_PAGECARD,
	    PCCARD_CIS_SOCKET_PAGECARD,
	    0,
	    "Socket Communications",
	    "Socket Communications PageCard"	},
	},
	{
	    PCCARD_VENDOR_SOCKET, PCCARD_PRODUCT_SOCKET_DUAL_RS232,
	    PCCARD_CIS_SOCKET_DUAL_RS232,
	    0,
	    "Socket Communications",
	    "Socket Communications Dual RS232"	},
	},
	{
	    PCCARD_VENDOR_SOCEKT, PCCARD_PRODUCT_SOCEKT_LP_ETHER_CF,
	    PCCARD_CIS_SOCEKT_LP_ETHER_CF,
	    0,
	    "",
	    "Socket Communications Low Power Ethernet CF"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_LAK_CD021BX,
	    PCCARD_CIS_TDK_LAK_CD021BX,
	    0,
	    "TDK Corporation",
	    "TDK LAK-CD021BX Ethernet"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_DFL9610,
	    PCCARD_CIS_TDK_DFL9610,
	    0,
	    "TDK Corporation",
	    "TDK DFL9610 Ethernet & Digital Cellular"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_LAK_CF010,
	    PCCARD_CIS_TDK_LAK_CF010,
	    0,
	    "TDK Corporation",
	    "TDK LAC-CF010"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_CE_10,
	    PCCARD_CIS_TDK_XIR_CE_10,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard Ethernet"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_CEM_10,
	    PCCARD_CIS_TDK_XIR_CEM_10,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard Ethernet + Modem"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_CEM_28,
	    PCCARD_CIS_TDK_XIR_CEM_28,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard Ethernet + Modem 28.8"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_PS_CE2_10,
	    PCCARD_CIS_TDK_XIR_PS_CE2_10,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard CE2 Ethernet"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_CE3_10_100,
	    PCCARD_CIS_TDK_XIR_CE3_10_100,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard CE3 10/100 Ethernet"	},
	},
	{
	    PCCARD_VENDOR_TDK, PCCARD_PRODUCT_TDK_XIR_CNW,
	    PCCARD_CIS_TDK_XIR_CNW,
	    0,
	    "TDK Corporation",
	    "Xircom CreditCard Netwave"	},
	},
	{
	    PCCARD_VENDOR_NEWMEDIA, PCCARD_PRODUCT_NEWMEDIA_BASICS,
	    PCCARD_CIS_NEWMEDIA_BASICS,
	    0,
	    "NewMedia Corporation",
	    "New Media BASICS Ethernet"	},
	},
	{
	    PCCARD_VENDOR_NEWMEDIA, PCCARD_PRODUCT_NEWMEDIA_BUSTOASTER,
	    PCCARD_CIS_NEWMEDIA_BUSTOASTER,
	    0,
	    "NewMedia Corporation",
	    "New Media BusToaster SCSI Host Adapter"	},
	},
	{
	    PCCARD_VENDOR_SMC, PCCARD_PRODUCT_SMC_8016,
	    PCCARD_CIS_SMC_8016,
	    0,
	    "Standard Microsystems Corporation",
	    "SMC 8016 EtherCard"	},
	},
	{
	    PCCARD_VENDOR_SMC, PCCARD_PRODUCT_SMC_EZCARD,
	    PCCARD_CIS_SMC_EZCARD,
	    0,
	    "Standard Microsystems Corporation",
	    "SMC EXCard 10 PCMCIA"	},
	},
	{
	    PCCARD_VENDOR_CONTEC, PCCARD_PRODUCT_CONTEC_CNETPC,
	    PCCARD_CIS_CONTEC_CNETPC,
	    0,
	    "Contec",
	    "Contec C-NET(PC)C"	},
	},
	{
	    PCCARD_VENDOR_ALLIEDTELESIS, PCCARD_PRODUCT_ALLIEDTELESIS_LA_PCM,
	    PCCARD_CIS_ALLIEDTELESIS_LA_PCM,
	    0,
	    "Allied Telesis K.K.",
	    "Allied Telesis LA-PCM"	},
	},
	{
	    PCCARD_VENDOR_RATOC, PCCARD_PRODUCT_RATOC_REX_R280,
	    PCCARD_CIS_RATOC_REX_R280,
	    0,
	    "RATOC System Inc.",
	    "RATOC REX-R280"	},
	},
	{
	    PCCARD_VENDOR_LUCENT, PCCARD_PRODUCT_LUCENT_WAVELAN_IEEE,
	    PCCARD_CIS_LUCENT_WAVELAN_IEEE,
	    0,
	    "Lucent Technologies",
	    "WaveLAN/IEEE"	},
	},
	{
	    PCCARD_VENDOR_BAY, PCCARD_PRODUCT_BAY_STACK_650,
	    PCCARD_CIS_BAY_STACK_650,
	    0,
	    "Bay Networks",
	    "BayStack 650 Wireless LAN"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_MEGAHERTZ_XJ2288,
	    PCCARD_CIS_MEGAHERTZ_XJ2288,
	    0,
	    "Megahertz Corporation",
	    "Megahertz XJ2288 Modem"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_PREMAX_PE200,
	    PCCARD_CIS_PREMAX_PE200,
	    0,
	    "Premax",
	    "PreMax PE-200"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_PLANET_SMARTCOM2000,
	    PCCARD_CIS_PLANET_SMARTCOM2000,
	    0,
	    "Planet",
	    "Planet SmartCOM 2000"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_PLANEX_FNW3600T,
	    PCCARD_CIS_PLANEX_FNW3600T,
	    0,
	    "Planex Communications Inc",
	    "Planex FNW-3600-T"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_DLINK_DE650,
	    PCCARD_CIS_DLINK_DE650,
	    0,
	    "D-Link",
	    "D-Link DE-650"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_DLINK_DE660,
	    PCCARD_CIS_DLINK_DE660,
	    0,
	    "D-Link",
	    "D-Link DE-660"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_RPTI_EP401,
	    PCCARD_CIS_RPTI_EP401,
	    0,
	    "RPTI",
	    "RPTI EP401"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_ACCTON_EN2212,
	    PCCARD_CIS_ACCTON_EN2212,
	    0,
	    "ACCTON",
	    "Accton EN2212"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_YEDATA_EXTERNAL_FDD,
	    PCCARD_CIS_YEDATA_EXTERNAL_FDD,
	    0,
	    "Y-E DATA",
	    "Y-E DATA External FDD"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_DIGITAL_DEPCMXX,
	    PCCARD_CIS_DIGITAL_DEPCMXX,
	    0,
	    "Digital Equipment Corporation",
	    "DEC DEPCM-BA"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_TEAC_IDECARDII,
	    PCCARD_CIS_TEAC_IDECARDII,
	    0,
	    "TEAC",
	    "TEAC IDE Card/II"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_LINKSYS_ECARD_2,
	    PCCARD_CIS_LINKSYS_ECARD_2,
	    0,
	    "Linksys Corporation",
	    "Linksys E-Card"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_COREGA_ETHER_PCC_T,
	    PCCARD_CIS_COREGA_ETHER_PCC_T,
	    0,
	    "Corega K.K.",
	    "Corega"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_COREGA_ETHER_II_PCC_T,
	    PCCARD_CIS_COREGA_ETHER_II_PCC_T,
	    0,
	    "Corega K.K.",
	    "Corega"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_COREGA_FAST_ETHER_PCC_TX,
	    PCCARD_CIS_COREGA_FAST_ETHER_PCC_TX,
	    0,
	    "Corega K.K.",
	    "Corega"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_SVEC_COMBOCARD,
	    PCCARD_CIS_SVEC_COMBOCARD,
	    0,
	    "SVEC/Hawking Technology",
	    "SVEC/Hawking Tech. Combo Card"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_SVEC_LANCARD,
	    PCCARD_CIS_SVEC_LANCARD,
	    0,
	    "SVEC/Hawking Technology",
	    "SVEC PCMCIA Lan Card"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_NAKAGAWAMETAL_LNT10TN,
	    PCCARD_CIS_NAKAGAWAMETAL_LNT10TN,
	    0,
	    "NAKAGAWA METAL",
	    "NAKAGAWA METAL LNT-10TN NE2000 Compatible Card"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_AMBICOM_AMB8002T,
	    PCCARD_CIS_AMBICOM_AMB8002T,
	    0,
	    "AmbiCom Inc",
	    "AmbiCom AMB8002T"	},
	},
	{
	    PCCARD_VENDOR_UNKNOWN, PCCARD_PRODUCT_IODATA_PCLAT,
	    PCCARD_CIS_IODATA_PCLAT,
	    0,
	    "I-O DATA",
	    "IO-DATA PCLA/T"	},
	},
	{
	    PCCARD_VENDOR_FUJITSU, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Fujitsu Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_SANDISK, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Sandisk Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_NEWMEDIA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "NewMedia Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_INTEL, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Intel Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_IBM, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "IBM Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_MOTOROLA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Motorola Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_3COM, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "3Com Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Megahertz Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_SOCKET, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Socket Communications",
	    NULL,
	},
	{
	    PCCARD_VENDOR_TDK, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "TDK Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_SMC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Standard Microsystems Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_USROBOTICS, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "US Robotics Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_MEGAHERTZ2, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Megahertz Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_ADAPTEC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Adaptec Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_COMPAQ1, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Compaq Corporation (1)",
	    NULL,
	},
	{
	    PCCARD_VENDOR_LINKSYS, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Linksys Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_SIMPLETECH, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Simple Technology",
	    NULL,
	},
	{
	    PCCARD_VENDOR_LUCENT, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Lucent Technologies",
	    NULL,
	},
	{
	    PCCARD_VENDOR_COMPAQ2, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Compaq Corporation (2)",
	    NULL,
	},
	{
	    PCCARD_VENDOR_DAYNA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Dayna Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_IODATA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "I-O DATA",
	    NULL,
	},
	{
	    PCCARD_VENDOR_BAY, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Bay Networks",
	    NULL,
	},
	{
	    PCCARD_VENDOR_COMPEX, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Compex Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_CONTEC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Contec",
	    NULL,
	},
	{
	    PCCARD_VENDOR_COREGA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Corega K.K.",
	    NULL,
	},
	{
	    PCCARD_VENDOR_ALLIEDTELESIS, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Allied Telesis K.K.",
	    NULL,
	},
	{
	    PCCARD_VENDOR_HAGIWARASYSCOM, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Hagiwara SYS-COM",
	    NULL,
	},
	{
	    PCCARD_VENDOR_RATOC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "RATOC System Inc.",
	    NULL,
	},
	{
	    PCCARD_VENDOR_PREMAX, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Premax",
	    NULL,
	},
	{
	    PCCARD_VENDOR_PLANET, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Planet",
	    NULL,
	},
	{
	    PCCARD_VENDOR_PLANEX, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Planex Communications Inc",
	    NULL,
	},
	{
	    PCCARD_VENDOR_DLINK, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "D-Link",
	    NULL,
	},
	{
	    PCCARD_VENDOR_RPTI, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "RPTI",
	    NULL,
	},
	{
	    PCCARD_VENDOR_ACCTON, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "ACCTON",
	    NULL,
	},
	{
	    PCCARD_VENDOR_YEDATA, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Y-E DATA",
	    NULL,
	},
	{
	    PCCARD_VENDOR_DIGITAL, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "Digital Equipment Corporation",
	    NULL,
	},
	{
	    PCCARD_VENDOR_TEAC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "TEAC",
	    NULL,
	},
	{
	    PCCARD_VENDOR_SVEC, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "SVEC/Hawking Technology",
	    NULL,
	},
	{
	    PCCARD_VENDOR_NAKAGAWAMETAL, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "NAKAGAWA METAL",
	    NULL,
	},
	{
	    PCCARD_VENDOR_AMBICOM, 0,
	    PCCARD_KNOWNDEV_NOPROD,
	    PCCARD_CIS_INVALID,
	    "AmbiCom Inc",
	    NULL,
	},
	{ 0, 0, { NULL, NULL, NULL, NULL }, 0, NULL, NULL, }
};
