/*	$FreeBSD: src/sys/dev/usb/usbdevs_data.h,v 1.32 2000/03/09 16:29:49 gehenna Exp $	*/

/*
 * THIS FILE IS AUTOMATICALLY GENERATED.  DO NOT EDIT.
 *
 * generated from:
 *	FreeBSD: src/sys/dev/usb/usbdevs,v 1.11 2000/03/09 16:28:58 gehenna Exp 
 */

/*
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (augustss@carlstedt.se) at
 * Carlstedt Research & Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

struct usb_knowndev usb_knowndevs[] = {
	{
	    USB_VENDOR_3COM, USB_PRODUCT_3COM_HOMECONN,
	    0,
	    "3Com Corp.",
	    "HomeConnect USB Camera",
	},
	{
	    USB_VENDOR_3COM, USB_PRODUCT_3COM_3C19250,
	    0,
	    "3Com Corp.",
	    "3C19250 Ethernet adapter",
	},
	{
	    USB_VENDOR_3COM, USB_PRODUCT_3COM_USR56K,
	    0,
	    "3Com Corp.",
	    "U.S.Robotics 56000 Voice Faxmodem Pro",
	},
	{
	    USB_VENDOR_3COMUSR, USB_PRODUCT_3COMUSR_USR56K,
	    0,
	    "U.S. Robotics",
	    "U.S.Robotics 56000 Voice Faxmodem Pro",
	},
	{
	    USB_VENDOR_ACER, USB_PRODUCT_ACER_ACERSCAN_C310U,
	    0,
	    "Acer Peripheral Inc.",
	    "Acerscan C310U",
	},
	{
	    USB_VENDOR_ACTIVEWIRE, USB_PRODUCT_ACTIVEWIRE_IOBOARD,
	    0,
	    "ActiveWire Inc.",
	    "I/O Board",
	},
	{
	    USB_VENDOR_ACTIVEWIRE, USB_PRODUCT_ACTIVEWIRE_IOBOARD_FW1,
	    0,
	    "ActiveWire Inc.",
	    "I/O Board, rev. 1 firmware",
	},
	{
	    USB_VENDOR_ADMTEK, USB_PRODUCT_ADMTEK_PEGASUS,
	    0,
	    "ADMtek Inc.",
	    "AN986 USB Ethernet adapter",
	},
	{
	    USB_VENDOR_ADS, USB_PRODUCT_ADS_UBS10BT,
	    0,
	    "ADS Technologies",
	    "UBS-10BT Ethernet adapter",
	},
	{
	    USB_VENDOR_ELECOM, USB_PRODUCT_ELECOM_MOUSE29UO,
	    0,
	    "Elecom Corp. Ltd.",
	    "mouse 29UO",
	},
	{
	    USB_VENDOR_AKS, USB_PRODUCT_AKS_USBHASP,
	    0,
	    "Fast Security AG",
	    "USB-HASP 0.06",
	},
	{
	    USB_VENDOR_ALTEC, USB_PRODUCT_ALTEC_ASC495,
	    0,
	    "Altec Lansing",
	    "ASC495 Speakers",
	},
	{
	    USB_VENDOR_APC, USB_PRODUCT_APC_UPSPRO500,
	    0,
	    "American Power Conversion",
	    "Back-UPS Pro 500",
	},
	{
	    USB_VENDOR_ANCHOR, USB_PRODUCT_ANCHOR_EZUSB,
	    0,
	    "Anchor Chips Inc.",
	    "EZUSB",
	},
	{
	    USB_VENDOR_AOX, USB_PRODUCT_AOX_USB101,
	    0,
	    "Aox Inc.",
	    "USB ethernet controller engine",
	},
	{
	    USB_VENDOR_ATEN, USB_PRODUCT_ATEN_UC1284,
	    0,
	    "ATEN International Corp. Ltd.",
	    "Parallel printer adapter",
	},
	{
	    USB_VENDOR_ATEN, USB_PRODUCT_ATEN_UC10T,
	    0,
	    "ATEN International Corp. Ltd.",
	    "10Mbps ethernet adapter",
	},
	{
	    USB_VENDOR_BILLIONTON, USB_PRODUCT_BILLIONTON_USB100,
	    0,
	    "Billionton Systems Inc.",
	    "USB100N 10/100 FastEthernet Adapter",
	},
	{
	    USB_VENDOR_BROTHER, USB_PRODUCT_BROTHER_HL1050,
	    0,
	    "Brother Industries Corp.",
	    "HL-1050 laser printer",
	},
	{
	    USB_VENDOR_BTC, USB_PRODUCT_BTC_BTC7932,
	    0,
	    "Behavior Tech. Computer",
	    "Keyboard with mouse port",
	},
	{
	    USB_VENDOR_CANON, USB_PRODUCT_CANON_S10,
	    0,
	    "Canon Inc.",
	    "PowerShot S10",
	},
	{
	    USB_VENDOR_CATC, USB_PRODUCT_CATC_NETMATE,
	    0,
	    "Computer Access Technology Corp.",
	    "Netmate ethernet adapter",
	},
	{
	    USB_VENDOR_CATC, USB_PRODUCT_CATC_NETMATE2,
	    0,
	    "Computer Access Technology Corp.",
	    "Netmate2 ethernet adapter",
	},
	{
	    USB_VENDOR_CATC, USB_PRODUCT_CATC_CHIEF,
	    0,
	    "Computer Access Technology Corp.",
	    "USB Chief Bus & Protocol Analyzer",
	},
	{
	    USB_VENDOR_CATC, USB_PRODUCT_CATC_ANDROMEDA,
	    0,
	    "Computer Access Technology Corp.",
	    "Andromeda hub",
	},
	{
	    USB_VENDOR_CHERRY, USB_PRODUCT_CHERRY_MY3000KBD,
	    0,
	    "Cherry Mikroschalter GmbH",
	    "My3000 keyboard",
	},
	{
	    USB_VENDOR_CHERRY, USB_PRODUCT_CHERRY_MY3000HUB,
	    0,
	    "Cherry Mikroschalter GmbH",
	    "My3000 hub",
	},
	{
	    USB_VENDOR_CHIC, USB_PRODUCT_CHIC_MOUSE1,
	    0,
	    "Chic Technology",
	    "mouse",
	},
	{
	    USB_VENDOR_CHICONY, USB_PRODUCT_CHICONY_KB8933,
	    0,
	    "Chicony Electronics Co., Ltd.",
	    "KB-8933 keyboard",
	},
	{
	    USB_VENDOR_CONNECTIX, USB_PRODUCT_CONNECTIX_QUICKCAM,
	    0,
	    "Connectix Corp.",
	    "QuickCam",
	},
	{
	    USB_VENDOR_COREGA, USB_PRODUCT_COREGA_ETHER_USB_T,
	    0,
	    "Corega",
	    "Ether USB-T",
	},
	{
	    USB_VENDOR_COREGA, USB_PRODUCT_COREGA_FETHER_USB_TX,
	    0,
	    "Corega",
	    "FEther USB-TX",
	},
	{
	    USB_VENDOR_CYPRESS, USB_PRODUCT_CYPRESS_MOUSE,
	    0,
	    "Cypress Semiconductor",
	    "mouse",
	},
	{
	    USB_VENDOR_CYPRESS, USB_PRODUCT_CYPRESS_THERMO,
	    0,
	    "Cypress Semiconductor",
	    "thermometer",
	},
	{
	    USB_VENDOR_DLINK, USB_PRODUCT_DLINK_DSB650C,
	    0,
	    "D-Link Corp",
	    "10Mbps ethernet adapter",
	},
	{
	    USB_VENDOR_DLINK, USB_PRODUCT_DLINK_DSB650TX,
	    0,
	    "D-Link Corp",
	    "10/100 ethernet adapter",
	},
	{
	    USB_VENDOR_DLINK, USB_PRODUCT_DLINK_DSB650TX_PNA,
	    0,
	    "D-Link Corp",
	    "1/10/100 ethernet adapter",
	},
	{
	    USB_VENDOR_DALLAS, USB_PRODUCT_DALLAS_J6502,
	    0,
	    "Dallas Semiconductor",
	    "J-6502 speakers",
	},
	{
	    USB_VENDOR_DIAMOND, USB_PRODUCT_DIAMOND_RIO500USB,
	    0,
	    "Diamond",
	    "Rio 500 USB",
	},
	{
	    USB_VENDOR_EIZO, USB_PRODUCT_EIZO_HUB,
	    0,
	    "EIZO",
	    "hub",
	},
	{
	    USB_VENDOR_EIZO, USB_PRODUCT_EIZO_MONITOR,
	    0,
	    "EIZO",
	    "monitor",
	},
	{
	    USB_VENDOR_ELSA, USB_PRODUCT_ELSA_MODEM1,
	    0,
	    "ELSA Gmbh",
	    "ELSA Modem Board",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_1S,
	    0,
	    "Entrega",
	    "1S serial connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_2S,
	    0,
	    "Entrega",
	    "2S serial connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_1S25,
	    0,
	    "Entrega",
	    "1S25 serial connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_4S,
	    0,
	    "Entrega",
	    "4S serial connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_E45,
	    0,
	    "Entrega",
	    "E45 Ethernet adapter",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_CENTRONICS,
	    0,
	    "Entrega",
	    "Centronics connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_1S9,
	    0,
	    "Entrega",
	    "1S9 serial connector",
	},
	{
	    USB_VENDOR_ENTREGA, USB_PRODUCT_ENTREGA_EZUSB,
	    0,
	    "Entrega",
	    "EZ-USB",
	},
	{
	    USB_VENDOR_ETEK, USB_PRODUCT_ETEK_1COM,
	    0,
	    "e-TEK Labs",
	    "Serial port",
	},
	{
	    USB_VENDOR_EPSON, USB_PRODUCT_EPSON_PRINTER2,
	    0,
	    "Seiko Epson Corp.",
	    "ISD USB Smart Cable for Mac",
	},
	{
	    USB_VENDOR_EPSON, USB_PRODUCT_EPSON_PRINTER3,
	    0,
	    "Seiko Epson Corp.",
	    "ISD USB Smart Cable",
	},
	{
	    USB_VENDOR_GRAVIS, USB_PRODUCT_GRAVIS_GAMEPADPRO,
	    0,
	    "Advanced Gravis Computer Tech. Ltd.",
	    "GamePad Pro",
	},
	{
	    USB_VENDOR_HANDSPRING, USB_PRODUCT_HANDSPRING_VISOR,
	    0,
	    "Handspring Inc.",
	    "Handspring Visor",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_4100C,
	    0,
	    "Hewlett Packard",
	    "Scanjet 4100C",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_S20,
	    0,
	    "Hewlett Packard",
	    "Photosmart S20",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_5200,
	    0,
	    "Hewlett Packard",
	    "Scanjet 5200",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_6300C,
	    0,
	    "Hewlett Packard",
	    "Scanjet 6300C",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_970CSE,
	    0,
	    "Hewlett Packard",
	    "Deskjet 970Cse",
	},
	{
	    USB_VENDOR_HP, USB_PRODUCT_HP_P1100,
	    0,
	    "Hewlett Packard",
	    "Photosmart P1100",
	},
	{
	    USB_VENDOR_INSIDEOUT, USB_PRODUCT_INSIDEOUT_EDGEPORT4,
	    0,
	    "Inside Out Networks",
	    "EdgePort/4 serial ports",
	},
	{
	    USB_VENDOR_INSYSTEM, USB_PRODUCT_INSYSTEM_F5U002,
	    0,
	    "In-System Design",
	    "Parallel printer adapter",
	},
	{
	    USB_VENDOR_INSYSTEM, USB_PRODUCT_INSYSTEM_ISD110,
	    0,
	    "In-System Design",
	    "IDE adapter",
	},
	{
	    USB_VENDOR_INTEL, USB_PRODUCT_INTEL_TESTBOARD,
	    0,
	    "Intel",
	    "82930 test board",
	},
	{
	    USB_VENDOR_IOMEGA, USB_PRODUCT_IOMEGA_ZIP100,
	    0,
	    "Iomega Corp.",
	    "Zip 100",
	},
	{
	    USB_VENDOR_KAWATSU, USB_PRODUCT_KAWATSU_MH4000P,
	    0,
	    "Kawatsu Semiconductor, Inc.",
	    "MiniHub 4000P",
	},
	{
	    USB_VENDOR_KODAK, USB_PRODUCT_KODAK_DC260,
	    0,
	    "Eastman Kodak Corp.",
	    "Digital Science DC260",
	},
	{
	    USB_VENDOR_KODAK, USB_PRODUCT_KODAK_DC265,
	    0,
	    "Eastman Kodak Corp.",
	    "Digital Science DC265",
	},
	{
	    USB_VENDOR_KODAK, USB_PRODUCT_KODAK_DC290,
	    0,
	    "Eastman Kodak Corp.",
	    "Digital Science DC290",
	},
	{
	    USB_VENDOR_KODAK, USB_PRODUCT_KODAK_DC240,
	    0,
	    "Eastman Kodak Corp.",
	    "Digital Science DC240",
	},
	{
	    USB_VENDOR_KODAK, USB_PRODUCT_KODAK_DC280,
	    0,
	    "Eastman Kodak Corp.",
	    "Digital Science DC280",
	},
	{
	    USB_VENDOR_KONICA, USB_PRODUCT_KONICA_CAMERA,
	    0,
	    "Konica Corp.",
	    "Digital Color Camera",
	},
	{
	    USB_VENDOR_KYE, USB_PRODUCT_KYE_NICHE,
	    0,
	    "KYE Systems Corp.",
	    "Niche mouse",
	},
	{
	    USB_VENDOR_KYE, USB_PRODUCT_KYE_FLIGHT2000,
	    0,
	    "KYE Systems Corp.",
	    "Flight 2000 joystick",
	},
	{
	    USB_VENDOR_LEXMARK, USB_PRODUCT_LEXMARK_S2450,
	    0,
	    "Lexmark International Inc.",
	    "Optra S 2450",
	},
	{
	    USB_VENDOR_LINKSYS, USB_PRODUCT_LINKSYS_USB100TX,
	    0,
	    "Linksys Inc.",
	    "USB100TX Ethernet",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_M2452,
	    0,
	    "Logitech Inc.",
	    "M2452 keyboard",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_M4848,
	    0,
	    "Logitech Inc.",
	    "M4848 mouse",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_QUICKCAM,
	    0,
	    "Logitech Inc.",
	    "QuickCam",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_QUICKCAMPRO,
	    0,
	    "Logitech Inc.",
	    "QuickCam Pro",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_N48,
	    0,
	    "Logitech Inc.",
	    "N48 mouse",
	},
	{
	    USB_VENDOR_LOGITECH, USB_PRODUCT_LOGITECH_MBA47,
	    0,
	    "Logitech Inc.",
	    "M-BA47 mouse",
	},
	{
	    USB_VENDOR_LUCENT, USB_PRODUCT_LUCENT_EVALKIT,
	    0,
	    "Lucent",
	    "USS-720 evaluation kit",
	},
	{
	    USB_VENDOR_MACALLY, USB_PRODUCT_MACALLY_MOUSE1,
	    0,
	    "Macally",
	    "mouse",
	},
	{
	    USB_VENDOR_MELCO, USB_PRODUCT_MELCO_LUATX,
	    0,
	    "Melco Inc.",
	    "LU-ATX Ethernet",
	},
	{
	    USB_VENDOR_MICROSOFT, USB_PRODUCT_MICROSOFT_INTELLIMOUSE,
	    0,
	    "Microsoft",
	    "IntelliMouse",
	},
	{
	    USB_VENDOR_MICROSOFT, USB_PRODUCT_MICROSOFT_NATURALKBD,
	    0,
	    "Microsoft",
	    "Natural Keyboard Elite",
	},
	{
	    USB_VENDOR_MICROSOFT, USB_PRODUCT_MICROSOFT_DDS80,
	    0,
	    "Microsoft",
	    "Digital Sound System 80",
	},
	{
	    USB_VENDOR_MICROSOFT, USB_PRODUCT_MICROSOFT_SIDEWINDER,
	    0,
	    "Microsoft",
	    "Sidewinder Precision Racing Wheel",
	},
	{
	    USB_VENDOR_MIDIMAN, USB_PRODUCT_MIDIMAN_MIDISPORT2X2,
	    0,
	    "Midiman",
	    "Midisport 2x2",
	},
	{
	    USB_VENDOR_MOTOROLA, USB_PRODUCT_MOTOROLA_MC141555,
	    0,
	    "Motorola",
	    "MC141555 hub controller",
	},
	{
	    USB_VENDOR_MULTITECH, USB_PRODUCT_MULTITECH_ATLAS,
	    0,
	    "MultiTech",
	    "MT5634ZBA-USB modem",
	},
	{
	    USB_VENDOR_MUSTEK, USB_PRODUCT_MUSTEK_MDC800,
	    0,
	    "Mustek Systems Inc.",
	    "MDC-800 digital camera",
	},
	{
	    USB_VENDOR_NEC, USB_PRODUCT_NEC_HUB,
	    0,
	    "NEC",
	    "hub",
	},
	{
	    USB_VENDOR_NEC, USB_PRODUCT_NEC_HUB_B,
	    0,
	    "NEC",
	    "hub",
	},
	{
	    USB_VENDOR_NETCHIP, USB_PRODUCT_NETCHIP_TURBOCONNECT,
	    0,
	    "NetChip Technology",
	    "Turbo-Connect",
	},
	{
	    USB_VENDOR_NETGEAR, USB_PRODUCT_NETGEAR_EA101,
	    0,
	    "BayNETGEAR Inc.",
	    "Ethernet adapter",
	},
	{
	    USB_VENDOR_OMNIVISION, USB_PRODUCT_OMNIVISION_OV511,
	    0,
	    "OmniVision",
	    "OV511 Camera",
	},
	{
	    USB_VENDOR_PERACOM, USB_PRODUCT_PERACOM_SERIAL1,
	    0,
	    "Peracom Networks Inc.",
	    "Serial Converter",
	},
	{
	    USB_VENDOR_PERACOM, USB_PRODUCT_PERACOM_ENET,
	    0,
	    "Peracom Networks Inc.",
	    "Ethernet adapter",
	},
	{
	    USB_VENDOR_PERACOM, USB_PRODUCT_PERACOM_ENET2,
	    0,
	    "Peracom Networks Inc.",
	    "Ethernet adapter",
	},
	{
	    USB_VENDOR_PHILIPS, USB_PRODUCT_PHILIPS_DSS350,
	    0,
	    "Philips",
	    "DSS 350 Digital Speaker System",
	},
	{
	    USB_VENDOR_PHILIPS, USB_PRODUCT_PHILIPS_DSS,
	    0,
	    "Philips",
	    "DSS XXX Digital Speaker System",
	},
	{
	    USB_VENDOR_PHILIPS, USB_PRODUCT_PHILIPS_HUB,
	    0,
	    "Philips",
	    "hub",
	},
	{
	    USB_VENDOR_PHILIPS, USB_PRODUCT_PHILIPS_DSS150,
	    0,
	    "Philips",
	    "DSS XXX Digital Speaker System",
	},
	{
	    USB_VENDOR_PIENGINEERING, USB_PRODUCT_PIENGINEERING_PS2USB,
	    0,
	    "P.I. Engineering",
	    "PS2 to Mac USB Adapter",
	},
	{
	    USB_VENDOR_PLX, USB_PRODUCT_PLX_TESTBOARD,
	    0,
	    "PLX",
	    "test board",
	},
	{
	    USB_VENDOR_PRIMAX, USB_PRODUCT_PRIMAX_COMFORT,
	    0,
	    "Primax Electronics",
	    "Comfort",
	},
	{
	    USB_VENDOR_PRIMAX, USB_PRODUCT_PRIMAX_MOUSEINABOX,
	    0,
	    "Primax Electronics",
	    "Mouse-in-a-Box",
	},
	{
	    USB_VENDOR_ROCKFIRE, USB_PRODUCT_ROCKFIRE_GAMEPAD,
	    0,
	    "Rockfire",
	    "gamepad 203USB",
	},
	{
	    USB_VENDOR_QTRONIX, USB_PRODUCT_QTRONIX_980N,
	    0,
	    "Qtronix Corp",
	    "Scorpion-980N keyboard",
	},
	{
	    USB_VENDOR_SANDISK, USB_PRODUCT_SANDISK_IMAGEMATE,
	    0,
	    "SanDisk Corp",
	    "USB ImageMate",
	},
	{
	    USB_VENDOR_SHUTTLE, USB_PRODUCT_SHUTTLE_EUSB,
	    0,
	    "Shuttle Technology",
	    "E-USB Bridge",
	},
	{
	    USB_VENDOR_SIIG, USB_PRODUCT_SIIG_DIGIFILMREADER,
	    0,
	    "SIIG",
	    "DigiFilm-Combo Reader",
	},
	{
	    USB_VENDOR_SIRIUS, USB_PRODUCT_SIRIUS_ROADSTER,
	    0,
	    "Sirius Technologies",
	    "NetComm Roadster II 56 USB",
	},
	{
	    USB_VENDOR_SMC, USB_PRODUCT_SMC_2102USB,
	    0,
	    "Standard Microsystems Corp",
	    "10Mbps ethernet adapter",
	},
	{
	    USB_VENDOR_SMC, USB_PRODUCT_SMC_2202USB,
	    0,
	    "Standard Microsystems Corp",
	    "10/100 ethernet adapter",
	},
	{
	    USB_VENDOR_SOLIDYEAR, USB_PRODUCT_SOLIDYEAR_KEYBOARD,
	    0,
	    "Solid Year",
	    "Solid Year USB keyboard",
	},
	{
	    USB_VENDOR_STMICRO, USB_PRODUCT_STMICRO_COMMUNICATOR,
	    0,
	    "STMicroelectronics",
	    "USB Communicator",
	},
	{
	    USB_VENDOR_SUN, USB_PRODUCT_SUN_KEYBOARD,
	    0,
	    "Sun Microsystems",
	    "Type 6 USB",
	},
	{
	    USB_VENDOR_TELEX, USB_PRODUCT_TELEX_MIC1,
	    0,
	    "Telex Communications Inc.",
	    "Enhanced USB Microphone",
	},
	{
	    USB_VENDOR_TI, USB_PRODUCT_TI_UTUSB41,
	    0,
	    "Texas Instruments",
	    "UT-USB41 hub",
	},
	{
	    USB_VENDOR_THRUST, USB_PRODUCT_THRUST_FUSION_PAD,
	    0,
	    "Thrustmaster",
	    "Fusion Digital Gamepad",
	},
	{
	    USB_VENDOR_UNIACCESS, USB_PRODUCT_UNIACCESS_PANACHE,
	    0,
	    "Universal Access",
	    "Panache Surf USB ISDN Adapter",
	},
	{
	    USB_VENDOR_VISION, USB_PRODUCT_VISION_VC6452V002,
	    0,
	    "VLSI Vision Ltd.",
	    "VC6452V002 Camera",
	},
	{
	    USB_VENDOR_WACOM, USB_PRODUCT_WACOM_CT0405U,
	    0,
	    "WACOM Corp. Ltd.",
	    "CT-0405-U Tablet",
	},
	{
	    USB_VENDOR_ZOOM, USB_PRODUCT_ZOOM_2986L,
	    0,
	    "Zoom Telephonics Inc.",
	    "2986L Fax modem",
	},
	{
	    USB_VENDOR_AOX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Aox Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_HP, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Hewlett Packard",
	    NULL,
	},
	{
	    USB_VENDOR_NEC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "NEC",
	    NULL,
	},
	{
	    USB_VENDOR_KODAK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Eastman Kodak Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_MELCO, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Melco Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_CATC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Computer Access Technology Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_GRAVIS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Advanced Gravis Computer Tech. Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_SUN, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Sun Microsystems",
	    NULL,
	},
	{
	    USB_VENDOR_LEXMARK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Lexmark International Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_NANAO, 0,
	    USB_KNOWNDEV_NOPROD,
	    "NANAO Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_THRUST, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Thrustmaster",
	    NULL,
	},
	{
	    USB_VENDOR_TI, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Texas Instruments",
	    NULL,
	},
	{
	    USB_VENDOR_KYE, 0,
	    USB_KNOWNDEV_NOPROD,
	    "KYE Systems Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_MICROSOFT, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Microsoft",
	    NULL,
	},
	{
	    USB_VENDOR_PRIMAX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Primax Electronics",
	    NULL,
	},
	{
	    USB_VENDOR_CHERRY, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Cherry Mikroschalter GmbH",
	    NULL,
	},
	{
	    USB_VENDOR_LOGITECH, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Logitech Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_BTC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Behavior Tech. Computer",
	    NULL,
	},
	{
	    USB_VENDOR_PHILIPS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Philips",
	    NULL,
	},
	{
	    USB_VENDOR_CONNECTIX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Connectix Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_LUCENT, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Lucent",
	    NULL,
	},
	{
	    USB_VENDOR_STMICRO, 0,
	    USB_KNOWNDEV_NOPROD,
	    "STMicroelectronics",
	    NULL,
	},
	{
	    USB_VENDOR_ACER, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Acer Peripheral Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_CANON, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Canon Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_CYPRESS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Cypress Semiconductor",
	    NULL,
	},
	{
	    USB_VENDOR_EPSON, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Seiko Epson Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_3COMUSR, 0,
	    USB_KNOWNDEV_NOPROD,
	    "U.S. Robotics",
	    NULL,
	},
	{
	    USB_VENDOR_KONICA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Konica Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_ALTEC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Altec Lansing",
	    NULL,
	},
	{
	    USB_VENDOR_SHUTTLE, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Shuttle Technology",
	    NULL,
	},
	{
	    USB_VENDOR_CHICONY, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Chicony Electronics Co., Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_BROTHER, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Brother Industries Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_DALLAS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Dallas Semiconductor",
	    NULL,
	},
	{
	    USB_VENDOR_3COM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "3Com Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_BELKIN, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Belkin Components",
	    NULL,
	},
	{
	    USB_VENDOR_KAWATSU, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Kawatsu Semiconductor, Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_APC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "American Power Conversion",
	    NULL,
	},
	{
	    USB_VENDOR_NETCHIP, 0,
	    USB_KNOWNDEV_NOPROD,
	    "NetChip Technology",
	    NULL,
	},
	{
	    USB_VENDOR_AKS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Fast Security AG",
	    NULL,
	},
	{
	    USB_VENDOR_UNIACCESS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Universal Access",
	    NULL,
	},
	{
	    USB_VENDOR_ANCHOR, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Anchor Chips Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_VISION, 0,
	    USB_KNOWNDEV_NOPROD,
	    "VLSI Vision Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_ATEN, 0,
	    USB_KNOWNDEV_NOPROD,
	    "ATEN International Corp. Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_MUSTEK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Mustek Systems Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_TELEX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Telex Communications Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_PERACOM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Peracom Networks Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_WACOM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "WACOM Corp. Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_ETEK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "e-TEK Labs",
	    NULL,
	},
	{
	    USB_VENDOR_EIZO, 0,
	    USB_KNOWNDEV_NOPROD,
	    "EIZO",
	    NULL,
	},
	{
	    USB_VENDOR_ELECOM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Elecom Corp. Ltd.",
	    NULL,
	},
	{
	    USB_VENDOR_ROCKFIRE, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Rockfire",
	    NULL,
	},
	{
	    USB_VENDOR_IOMEGA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Iomega Corp.",
	    NULL,
	},
	{
	    USB_VENDOR_OMNIVISION, 0,
	    USB_KNOWNDEV_NOPROD,
	    "OmniVision",
	    NULL,
	},
	{
	    USB_VENDOR_INSYSTEM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "In-System Design",
	    NULL,
	},
	{
	    USB_VENDOR_APPLE, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Apple Computer",
	    NULL,
	},
	{
	    USB_VENDOR_QTRONIX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Qtronix Corp",
	    NULL,
	},
	{
	    USB_VENDOR_ELSA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "ELSA Gmbh",
	    NULL,
	},
	{
	    USB_VENDOR_EIZONANAO, 0,
	    USB_KNOWNDEV_NOPROD,
	    "EIZO Nanao",
	    NULL,
	},
	{
	    USB_VENDOR_KLSI, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Kawasaki LSI",
	    NULL,
	},
	{
	    USB_VENDOR_PIENGINEERING, 0,
	    USB_KNOWNDEV_NOPROD,
	    "P.I. Engineering",
	    NULL,
	},
	{
	    USB_VENDOR_CHIC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Chic Technology",
	    NULL,
	},
	{
	    USB_VENDOR_SOLIDYEAR, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Solid Year",
	    NULL,
	},
	{
	    USB_VENDOR_MACALLY, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Macally",
	    NULL,
	},
	{
	    USB_VENDOR_LINKSYS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Linksys Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_MULTITECH, 0,
	    USB_KNOWNDEV_NOPROD,
	    "MultiTech",
	    NULL,
	},
	{
	    USB_VENDOR_ADS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "ADS Technologies",
	    NULL,
	},
	{
	    USB_VENDOR_SIRIUS, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Sirius Technologies",
	    NULL,
	},
	{
	    USB_VENDOR_SMC, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Standard Microsystems Corp",
	    NULL,
	},
	{
	    USB_VENDOR_MIDIMAN, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Midiman",
	    NULL,
	},
	{
	    USB_VENDOR_SANDISK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "SanDisk Corp",
	    NULL,
	},
	{
	    USB_VENDOR_ADMTEK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "ADMtek Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_COREGA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Corega",
	    NULL,
	},
	{
	    USB_VENDOR_SIIG, 0,
	    USB_KNOWNDEV_NOPROD,
	    "SIIG",
	    NULL,
	},
	{
	    USB_VENDOR_ZOOM, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Zoom Telephonics Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_HANDSPRING, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Handspring Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_DIAMOND, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Diamond",
	    NULL,
	},
	{
	    USB_VENDOR_NETGEAR, 0,
	    USB_KNOWNDEV_NOPROD,
	    "BayNETGEAR Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_ACTIVEWIRE, 0,
	    USB_KNOWNDEV_NOPROD,
	    "ActiveWire Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_BILLIONTON, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Billionton Systems Inc.",
	    NULL,
	},
	{
	    USB_VENDOR_MOTOROLA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Motorola",
	    NULL,
	},
	{
	    USB_VENDOR_PLX, 0,
	    USB_KNOWNDEV_NOPROD,
	    "PLX",
	    NULL,
	},
	{
	    USB_VENDOR_INSIDEOUT, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Inside Out Networks",
	    NULL,
	},
	{
	    USB_VENDOR_ENTREGA, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Entrega",
	    NULL,
	},
	{
	    USB_VENDOR_DLINK, 0,
	    USB_KNOWNDEV_NOPROD,
	    "D-Link Corp",
	    NULL,
	},
	{
	    USB_VENDOR_INTEL, 0,
	    USB_KNOWNDEV_NOPROD,
	    "Intel",
	    NULL,
	},
	{ 0, 0, 0, NULL, NULL, }
};
