/* 
 * static configuration for libncp
 *
 * $FreeBSD: src/sys/netncp/ncp_cfg.h,v 1.1 1999/10/12 11:56:38 bp Exp $
 */

#define NCP_NLS_KOI2CP866
#define NCP_NLS_DEFAULT NCP_NLS_KOI_866
#define NCP_PREFIX	""
