/* $FreeBSD: src/sys/alpha/tc/tcdevs.h,v 1.2 1999/08/28 00:39:11 peter Exp $ */
/*	$NetBSD: tcdevs.h,v 1.12 1998/06/18 23:29:46 jonathan Exp $	*/

/*
 * THIS FILE AUTOMATICALLY GENERATED.  DO NOT EDIT.
 *
 * generated from:
 *	NetBSD: tcdevs,v 1.10 1998/06/18 23:28:12 jonathan Exp 
 */

/*
 * Copyright (c) 1996 Christopher G. Demetriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Christopher G. Demetriou
 *	for the NetBSD Project.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define	TC_DEVICE_AV01B_AA	"lofi"
#define	TC_DESCRIPTION_AV01B_AA	"DecAudio \"LoFi\" audio/isdn"

#define	TC_DEVICE_DGLTA_FA	"otto"
#define	TC_DESCRIPTION_DGLTA_FA	"DGLTA ATM"

#define	TC_DEVICE_FORE_ATM	"fa"
#define	TC_DESCRIPTION_FORE_ATM	"Fore TCA-100 ATM"

#define	TC_DEVICE_KZTSA_AA	"tza"
#define	TC_DESCRIPTION_KZTSA_AA	"TZA FWD SCSI"

#define	TC_DEVICE_OTTO	"otto"
#define	TC_DESCRIPTION_OTTO	"DEC SRC \"OTTO\" ATM"

#define	TC_DEVICE_PMAD_AA	"le"
#define	TC_DESCRIPTION_PMAD_AA	"LANCE Ethernet"

#define	TC_DEVICE_PMAF_AA	"fza"
#define	TC_DESCRIPTION_PMAF_AA	"DEC FDDIcontroller 700 (DEFZA; fiber optic)"

#define	TC_DEVICE_PMAF_F	"fta"
#define	TC_DESCRIPTION_PMAF_F	"DEFTA FDDI"

#define	TC_DEVICE_PMAG_AA	"mfb"
#define	TC_DESCRIPTION_PMAG_AA	"Monochrome Frame Buffer"

#define	TC_DEVICE_PMAG_BA	"cfb"
#define	TC_DESCRIPTION_PMAG_BA	"Color Frame Buffer"

#define	TC_DEVICE_PMAG_CA	"ga"
#define	TC_DESCRIPTION_PMAG_CA	"2D Graphics"

#define	TC_DEVICE_PMAG_DA	"gq"
#define	TC_DESCRIPTION_PMAG_DA	"3D Graphics (LM)"

#define	TC_DEVICE_PMAG_DV	"xcfb"
#define	TC_DESCRIPTION_PMAG_DV	"Maxine Color Frame Buffer"

#define	TC_DEVICE_PMAG_FA	"gq"
#define	TC_DESCRIPTION_PMAG_FA	"3D Graphics (HE)"

#define	TC_DEVICE_PMAGB_BA	"sfb"
#define	TC_DESCRIPTION_PMAGB_BA	"Smart Frame Buffer"

#define	TC_DEVICE_PMAGD	"sfbp"
#define	TC_DESCRIPTION_PMAGD	"Smart Frame Buffer Plus, unknown bpp"

#define	TC_DEVICE_PMAGD_AA	"sfbp"
#define	TC_DESCRIPTION_PMAGD_AA	"Smart Frame Buffer Plus, 8bpp"

#define	TC_DEVICE_PMAGD_BA	"sfbp"
#define	TC_DESCRIPTION_PMAGD_BA	"Smart Frame Buffer Plus, 32bpp"

#define	TC_DEVICE_PMAZ_AA	"asc"
#define	TC_DESCRIPTION_PMAZ_AA	"53c94 SCSI"

#define	TC_DEVICE_PMAZ_DS	"tcds"
#define	TC_DESCRIPTION_PMAZ_DS	"53c94 TCDS SCSI (baseboard)"

#define	TC_DEVICE_PMAZ_FS	"tcds"
#define	TC_DESCRIPTION_PMAZ_FS	"53c94 TCDS Fast SCSI (baseboard)"

#define	TC_DEVICE_PMAZB_AA	"tcds"
#define	TC_DESCRIPTION_PMAZB_AA	"53c94 TCDS SCSI option card"

#define	TC_DEVICE_PMAZC_AA	"tcds"
#define	TC_DESCRIPTION_PMAZC_AA	"53c94 TCDS Fast SCSI option card"

#define	TC_DEVICE_T1D4PKT	"ds"
#define	TC_DESCRIPTION_T1D4PKT	"DECWRL Turbochannel T1"

#define	TC_DEVICE_T3PKT	"tt"
#define	TC_DESCRIPTION_T3PKT	"DECWRL Turbochannel T3"

#define	TC_DEVICE_PMAT_AA	"tra"
#define	TC_DESCRIPTION_PMAT_AA	"DEC TurboChannel Token Ring Controller"
