/*	$FreeBSD: src/sys/dev/pccard/pccarddevs.h,v 1.3 1999/11/29 20:54:14 imp Exp $	*/

/*
 * THIS FILE AUTOMATICALLY GENERATED.  DO NOT EDIT.
 *
 * generated from:
 *	FreeBSD: src/sys/dev/pccard/pccarddevs,v 1.3 1999/11/29 20:53:36 imp Exp 
 */
/* $NetBSD: pcmciadevs,v 1.60 1999/11/29 02:28:19 jun Exp $ */

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Christos Zoulas.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * List of known PCMCIA vendors
 */

#define	PCCARD_VENDOR_FUJITSU	0x0004	/* Fujitsu Corporation */
#define	PCCARD_VENDOR_SANDISK	0x0045	/* Sandisk Corporation */
#define	PCCARD_VENDOR_NEWMEDIA	0x0057	/* NewMedia Corporation */
#define	PCCARD_VENDOR_INTEL	0x0089	/* Intel Corporation */
#define	PCCARD_VENDOR_IBM	0x00a4	/* IBM Corporation */
#define	PCCARD_VENDOR_MOTOROLA	0x0109	/* Motorola Corporation */
#define	PCCARD_VENDOR_3COM	0x0101	/* 3Com Corporation */
#define	PCCARD_VENDOR_MEGAHERTZ	0x0102	/* Megahertz Corporation */
#define	PCCARD_VENDOR_SOCKET	0x0104	/* Socket Communications */
#define	PCCARD_VENDOR_TDK	0x0105	/* TDK Corporation */
#define	PCCARD_VENDOR_SMC	0x0108	/* Standard Microsystems Corporation */
#define	PCCARD_VENDOR_USROBOTICS	0x0115	/* US Robotics Corporation */
#define	PCCARD_VENDOR_MEGAHERTZ2	0x0128	/* Megahertz Corporation */
#define	PCCARD_VENDOR_ADAPTEC	0x012f	/* Adaptec Corporation */
#define	PCCARD_VENDOR_COMPAQ1	0x0138	/* Compaq Corporation (1) */
#define	PCCARD_VENDOR_LINKSYS	0x0149	/* Linksys Corporation */
#define	PCCARD_VENDOR_SIMPLETECH	0x014d	/* Simple Technology */
#define	PCCARD_VENDOR_LUCENT	0x0156	/* Lucent Technologies */
#define	PCCARD_VENDOR_COMPAQ2	0x0183	/* Compaq Corporation (2) */
#define	PCCARD_VENDOR_DAYNA	0x0194	/* Dayna Corporation */
#define	PCCARD_VENDOR_IODATA	0x01bf	/* I-O DATA */
#define	PCCARD_VENDOR_BAY	0x01eb	/* Bay Networks */
#define	PCCARD_VENDOR_COMPEX	0x8a01	/* Compex Corporation */
#define	PCCARD_VENDOR_CONTEC	0xc001	/* Contec */
#define	PCCARD_VENDOR_COREGA	0xc00f	/* Corega K.K. */
#define	PCCARD_VENDOR_ALLIEDTELESIS	0xc00f	/* Allied Telesis K.K. */
#define	PCCARD_VENDOR_HAGIWARASYSCOM	0xc012	/* Hagiwara SYS-COM */
#define	PCCARD_VENDOR_RATOC	0xc015	/* RATOC System Inc. */

/*
 * List of known products.  Grouped by vendor.
 */

/* Adaptec Products */
#define	PCCARD_CIS_ADAPTEC_APA1460_1	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_ADAPTEC_APA1460_1	0x0001
#define	PCCARD_STR_ADAPTEC_APA1460_1	"Adaptec APA-1460/A SCSI Host Adapter"
#define	PCCARD_CIS_ADAPTEC_APA1460_2	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_ADAPTEC_APA1460_2	0x0002
#define	PCCARD_STR_ADAPTEC_APA1460_2	"Adaptec APA-1460/B SCSI Host Adapter"

/* 3COM Products */
#define	PCCARD_CIS_3COM_3C562	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3C562	0x0562
#define	PCCARD_STR_3COM_3C562	"3Com 3c562 33.6 Modem/10Mbps Ethernet"
#define	PCCARD_CIS_3COM_3C589	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3C589	0x0589
#define	PCCARD_STR_3COM_3C589	"3Com 3c589 10Mbps Ethernet"
#define	PCCARD_CIS_3COM_3C574	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3C574	0x0574
#define	PCCARD_STR_3COM_3C574	"3Com 3c574-TX 10/100Mbps Ethernet"
#define	PCCARD_CIS_3COM_3CXM056BNW	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3CXM056BNW	0x002f
#define	PCCARD_STR_3COM_3CXM056BNW	"3Com/NoteWorthy 3CXM056-BNW 56K Modem"
#define	PCCARD_CIS_3COM_3CXEM556	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3CXEM556	0x0035
#define	PCCARD_STR_3COM_3CXEM556	"3Com/Megahertz 3CXEM556 Ethernet/Modem"
#define	PCCARD_CIS_3COM_3CXEM556INT	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3CXEM556INT	0x003d
#define	PCCARD_STR_3COM_3CXEM556INT	"3Com/Megahertz 3CXEM556-INT Ethernet/Modem"
#define	PCCARD_CIS_3COM_3CCFEM556BI	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_3COM_3CCFEM556BI	0x0556
#define	PCCARD_STR_3COM_3CCFEM556BI	"3Com/Megahertz 3CCFEM556BI Ethernet/Modem"

/* Compex Products */
#define	PCCARD_CIS_COMPEX_LINKPORT_ENET_B	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_COMPEX_LINKPORT_ENET_B	0x0100
#define	PCCARD_STR_COMPEX_LINKPORT_ENET_B	"Compex Linkport ENET-B Ethernet"

/* Lasat Products */
#define	PCCARD_CIS_LASAT_CREDIT_288	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_LASAT_CREDIT_288	0x2811
#define	PCCARD_STR_LASAT_CREDIT_288	"Lasat Credit 288 Modem"

/* Dayna Products */
#define	PCCARD_CIS_DAYNA_COMMUNICARD_E_1	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_DAYNA_COMMUNICARD_E_1	0x002d
#define	PCCARD_STR_DAYNA_COMMUNICARD_E_1	"Dayna CommuniCard E"
#define	PCCARD_CIS_DAYNA_COMMUNICARD_E_2	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_DAYNA_COMMUNICARD_E_2	0x002f
#define	PCCARD_STR_DAYNA_COMMUNICARD_E_2	"Dayna CommuniCard E"

/* DIGITAL Products */
#define	PCCARD_CIS_DIGITAL_MOBILE_MEDIA_CDROM	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_DIGITAL_MOBILE_MEDIA_CDROM	0x0d00
#define	PCCARD_STR_DIGITAL_MOBILE_MEDIA_CDROM	"Digital Mobile Media CD-ROM"

/* Fujutsu Products */
#define	PCCARD_CIS_FUJITSU_SCSI600	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_FUJITSU_SCSI600	0x0401
#define	PCCARD_STR_FUJITSU_SCSI600	"Fujitsu SCSI 600 Interface"

/* Motorola Products */
#define	PCCARD_CIS_MOTOROLA_POWER144	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MOTOROLA_POWER144	0x0105
#define	PCCARD_STR_MOTOROLA_POWER144	"Motorola Power 14.4 Modem"
#define	PCCARD_CIS_MOTOROLA_PM100C	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MOTOROLA_PM100C	0x0302
#define	PCCARD_STR_MOTOROLA_PM100C	"Motorola Personal Messenger 100C CDPD Modem"

/* Fujitsu Products */
#define	PCCARD_CIS_FUJITSU_LA501	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_FUJITSU_LA501	0x2000
#define	PCCARD_STR_FUJITSU_LA501	"Fujitsu Towa LA501 Ethernet"
#define	PCCARD_CIS_FUJITSU_LA10S	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_FUJITSU_LA10S	0x1003
#define	PCCARD_STR_FUJITSU_LA10S	"Fujitsu Compact Flash Ethernet"

/* IBM Products */
#define	PCCARD_CIS_IBM_3270	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_3270	0x0001
#define	PCCARD_STR_IBM_3270	"3270 Emulation"
#define	PCCARD_CIS_IBM_INFOMOVER	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_INFOMOVER	0x0002
#define	PCCARD_STR_IBM_INFOMOVER	"National Semiconductor InfoMover"
#define	PCCARD_CIS_IBM_5250	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_5250	0x000b
#define	PCCARD_STR_IBM_5250	"5250 Emulation"
#define	PCCARD_CIS_IBM_TROPIC	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_TROPIC	0x001e
#define	PCCARD_STR_IBM_TROPIC	"Token Ring 4/16"
#define	PCCARD_CIS_IBM_HOME_AND_AWAY	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_HOME_AND_AWAY	0x002e
#define	PCCARD_STR_IBM_HOME_AND_AWAY	"IBM Home and Away Modem"
#define	PCCARD_CIS_IBM_WIRELESS_LAN_ENTRY	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_WIRELESS_LAN_ENTRY	0x0032
#define	PCCARD_STR_IBM_WIRELESS_LAN_ENTRY	"Wireless LAN Entry"
#define	PCCARD_CIS_IBM_PORTABLE_CDROM	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_PORTABLE_CDROM	0x002d
#define	PCCARD_STR_IBM_PORTABLE_CDROM	"PCMCIA Portable CD-ROM Drive"
#define	PCCARD_CIS_IBM_ETHERJET	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IBM_ETHERJET	0x003f
#define	PCCARD_STR_IBM_ETHERJET	"IBM EtherJet Ethernet"

/* I-O DATA */
#define	PCCARD_CIS_IODATA_PCLATE	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_IODATA_PCLATE	0x2216
#define	PCCARD_STR_IODATA_PCLATE	"I-O DATA PCLA/TE"

/* Linksys corporation */
#define	PCCARD_CIS_LINKSYS_ECARD_1	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_LINKSYS_ECARD_1	0x0265
#define	PCCARD_STR_LINKSYS_ECARD_1	"Linksys EthernetCard or D-Link DE-650"
#define	PCCARD_CIS_LINKSYS_COMBO_ECARD	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_LINKSYS_COMBO_ECARD	0xc1ab
#define	PCCARD_STR_LINKSYS_COMBO_ECARD	"Linksys Combo EthernetCard"
#define	PCCARD_CIS_LINKSYS_TRUST_COMBO_ECARD	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_LINKSYS_TRUST_COMBO_ECARD	0x021b
#define	PCCARD_STR_LINKSYS_TRUST_COMBO_ECARD	"Trust (Linksys) Combo EthernetCard"

/* Megahertz Products */
#define	PCCARD_CIS_MEGAHERTZ_XJ4288	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ_XJ4288	0x0023
#define	PCCARD_STR_MEGAHERTZ_XJ4288	"Megahertz XJ4288 Modem"
#define	PCCARD_CIS_MEGAHERTZ_XJ4336	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ_XJ4336	0x0027
#define	PCCARD_STR_MEGAHERTZ_XJ4336	"Megahertz XJ4336 Modem"
#define	PCCARD_CIS_MEGAHERTZ_XJ5560	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ_XJ5560	0x0034
#define	PCCARD_STR_MEGAHERTZ_XJ5560	"Megahertz X-JACK 56kbps Modem"
#define	PCCARD_CIS_MEGAHERTZ2_XJACK	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ2_XJACK	0x0103
#define	PCCARD_STR_MEGAHERTZ2_XJACK	"Megahertz X-JACK Ethernet"
#define	PCCARD_CIS_MEGAHERTZ_XJEM3336	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ_XJEM3336	0x0006
#define	PCCARD_STR_MEGAHERTZ_XJEM3336	"Megahertz X-JACK Ethernet Modem"

/* US Robotics Products */
#define	PCCARD_CIS_USROBOTICS_WORLDPORT144	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_USROBOTICS_WORLDPORT144	0x3330
#define	PCCARD_STR_USROBOTICS_WORLDPORT144	"US Robotics WorldPort 14.4 Modem"

/* Sandisk Products */
#define	PCCARD_CIS_SANDISK_SDCFB	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SANDISK_SDCFB	0x0401
#define	PCCARD_STR_SANDISK_SDCFB	"Sandisk CompactFlash Card"

/* Simple Technology Products */
#define	PCCARD_CIS_SIMPLETECH_COMMUNICATOR288	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SIMPLETECH_COMMUNICATOR288	0x0100
#define	PCCARD_STR_SIMPLETECH_COMMUNICATOR288	"Simple Technology 28.8 Communicator"
/* Simpletech ID also used by Symbol */
#define	PCCARD_CIS_SIMPLETECH_SPECTRUM24	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SIMPLETECH_SPECTRUM24	0x801
#define	PCCARD_STR_SIMPLETECH_SPECTRUM24	"Symbol Spectrum24 WLAN Adapter"

/* Socket Communications Products */
#define	PCCARD_CIS_SOCKET_PAGECARD	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SOCKET_PAGECARD	0x0003
#define	PCCARD_STR_SOCKET_PAGECARD	"Socket Communications PageCard"
#define	PCCARD_CIS_SOCKET_DUAL_RS232	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SOCKET_DUAL_RS232	0x0006
#define	PCCARD_STR_SOCKET_DUAL_RS232	"Socket Communications Dual RS232"
#define	PCCARD_CIS_SOCEKT_LP_ETHER_CF	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SOCEKT_LP_ETHER_CF	0x0075
#define	PCCARD_STR_SOCEKT_LP_ETHER_CF	"Socket Communications Low Power Ethernet CF"

/* TDK Products */
#define	PCCARD_CIS_TDK_LAK_CD021BX	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_LAK_CD021BX	0x0200
#define	PCCARD_STR_TDK_LAK_CD021BX	"TDK LAK-CD021BX Ethernet"
#define	PCCARD_CIS_TDK_DFL9610	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_DFL9610	0x0d0a
#define	PCCARD_STR_TDK_DFL9610	"TDK DFL9610 Ethernet & Digital Cellular"
#define	PCCARD_CIS_TDK_LAK_CF010	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_LAK_CF010	0x0900
#define	PCCARD_STR_TDK_LAK_CF010	"TDK LAC-CF010"

/* TDK Vendor ID also used by Xircom! */
#define	PCCARD_CIS_TDK_XIR_CE_10	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_CE_10	0x0108
#define	PCCARD_STR_TDK_XIR_CE_10	"Xircom CreditCard Ethernet"
#define	PCCARD_CIS_TDK_XIR_CEM_10	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_CEM_10	0x110a
#define	PCCARD_STR_TDK_XIR_CEM_10	"Xircom CreditCard Ethernet + Modem"
#define	PCCARD_CIS_TDK_XIR_CEM_28	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_CEM_28	0x110b
#define	PCCARD_STR_TDK_XIR_CEM_28	"Xircom CreditCard Ethernet + Modem 28.8"
#define	PCCARD_CIS_TDK_XIR_PS_CE2_10	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_PS_CE2_10	0x010b
#define	PCCARD_STR_TDK_XIR_PS_CE2_10	"Xircom CreditCard CE2 Ethernet"
#define	PCCARD_CIS_TDK_XIR_CE3_10_100	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_CE3_10_100	0x010a
#define	PCCARD_STR_TDK_XIR_CE3_10_100	"Xircom CreditCard CE3 10/100 Ethernet"
#define	PCCARD_CIS_TDK_XIR_CNW	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_TDK_XIR_CNW	0x0802
#define	PCCARD_STR_TDK_XIR_CNW	"Xircom CreditCard Netwave"

/* New Media Products */
#define	PCCARD_CIS_NEWMEDIA_BASICS	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_NEWMEDIA_BASICS	0x0019
#define	PCCARD_STR_NEWMEDIA_BASICS	"New Media BASICS Ethernet"
#define	PCCARD_CIS_NEWMEDIA_BUSTOASTER	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_NEWMEDIA_BUSTOASTER	0xc102
#define	PCCARD_STR_NEWMEDIA_BUSTOASTER	"New Media BusToaster SCSI Host Adapter"

/* Standard Microsystems Corporation Products */
#define	PCCARD_CIS_SMC_8016	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SMC_8016	0x0105
#define	PCCARD_STR_SMC_8016	"SMC 8016 EtherCard"
#define	PCCARD_CIS_SMC_EZCARD	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_SMC_EZCARD	0x8022
#define	PCCARD_STR_SMC_EZCARD	"SMC EXCard 10 PCMCIA"

/* Contec C-NET(PC) */
#define	PCCARD_CIS_CONTEC_CNETPC	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_CONTEC_CNETPC	0x0000
#define	PCCARD_STR_CONTEC_CNETPC	"Contec C-NET(PC)C"

/* Allied Telesis K.K. */
#define	PCCARD_CIS_ALLIEDTELESIS_LA_PCM	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_ALLIEDTELESIS_LA_PCM	0x0002
#define	PCCARD_STR_ALLIEDTELESIS_LA_PCM	"Allied Telesis LA-PCM"

/* RATOC System Inc. Products */
#define	PCCARD_CIS_RATOC_REX_R280	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_RATOC_REX_R280	0x1
#define	PCCARD_STR_RATOC_REX_R280	"RATOC REX-R280"

/* Lucent WaveLAN/IEEE */
#define	PCCARD_CIS_LUCENT_WAVELAN_IEEE	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_LUCENT_WAVELAN_IEEE	0x0002
#define	PCCARD_STR_LUCENT_WAVELAN_IEEE	"WaveLAN/IEEE"

#define	PCCARD_CIS_BAY_STACK_650	{ NULL, NULL, NULL, NULL }
#define	PCCARD_PRODUCT_BAY_STACK_650	0x804
#define	PCCARD_STR_BAY_STACK_650	"BayStack 650 Wireless LAN"

/* Cards we know only by their cis */
#define	PCCARD_VENDOR_PREMAX	-1	/* Premax */
#define	PCCARD_VENDOR_PLANET	-1	/* Planet */
#define	PCCARD_VENDOR_PLANEX	-1	/* Planex Communications Inc */
#define	PCCARD_VENDOR_DLINK	-1	/* D-Link */
#define	PCCARD_VENDOR_RPTI	-1	/* RPTI */
#define	PCCARD_VENDOR_ACCTON	-1	/* ACCTON */
#define	PCCARD_VENDOR_YEDATA	-1	/* Y-E DATA */
#define	PCCARD_VENDOR_DIGITAL	-1	/* Digital Equipment Corporation */
#define	PCCARD_VENDOR_TEAC	-1	/* TEAC */
#define	PCCARD_VENDOR_SVEC	-1	/* SVEC/Hawking Technology */
#define	PCCARD_VENDOR_NAKAGAWAMETAL	-1	/* NAKAGAWA METAL */
#define	PCCARD_VENDOR_AMBICOM	-1	/* AmbiCom Inc */

#define	PCCARD_CIS_MEGAHERTZ_XJ2288	{ "MEGAHERTZ", "MODEM XJ2288", NULL, NULL }
#define	PCCARD_PRODUCT_MEGAHERTZ_XJ2288	-1
#define	PCCARD_STR_MEGAHERTZ_XJ2288	"Megahertz XJ2288 Modem"
#define	PCCARD_CIS_PREMAX_PE200	{ "PMX   ", "PE-200", NULL, NULL }
#define	PCCARD_PRODUCT_PREMAX_PE200	-1
#define	PCCARD_STR_PREMAX_PE200	"PreMax PE-200"
#define	PCCARD_CIS_PLANET_SMARTCOM2000	{ "PCMCIA", "UE2212", NULL, NULL }
#define	PCCARD_PRODUCT_PLANET_SMARTCOM2000	-1
#define	PCCARD_STR_PLANET_SMARTCOM2000	"Planet SmartCOM 2000"
/*
 * vendor ID of FNW-3600-T is LINKSYS(0x0149) and product ID is 0xc1ab, but
 * it conflicts with LINKSYS Combo EhternetCard.
 */
#define	PCCARD_CIS_PLANEX_FNW3600T	{ "Fast Ethernet", "Adapter", "1.0", NULL }
#define	PCCARD_PRODUCT_PLANEX_FNW3600T	-1
#define	PCCARD_STR_PLANEX_FNW3600T	"Planex FNW-3600-T"
#define	PCCARD_CIS_DLINK_DE650	{ "D-Link", "DE-650", NULL, NULL }
#define	PCCARD_PRODUCT_DLINK_DE650	-1
#define	PCCARD_STR_DLINK_DE650	"D-Link DE-650"
#define	PCCARD_CIS_DLINK_DE660	{ "D-Link", "DE-660", NULL, NULL }
#define	PCCARD_PRODUCT_DLINK_DE660	-1
#define	PCCARD_STR_DLINK_DE660	"D-Link DE-660"
#define	PCCARD_CIS_RPTI_EP401	{ "RPTI", "EP401 Ethernet NE2000 Compatible", NULL, NULL }
#define	PCCARD_PRODUCT_RPTI_EP401	-1
#define	PCCARD_STR_RPTI_EP401	"RPTI EP401"
#define	PCCARD_CIS_ACCTON_EN2212	{ "ACCTON", "EN2212", NULL, NULL }
#define	PCCARD_PRODUCT_ACCTON_EN2212	-1
#define	PCCARD_STR_ACCTON_EN2212	"Accton EN2212"
#define	PCCARD_CIS_YEDATA_EXTERNAL_FDD	{ "Y-E DATA", "External FDD", NULL, NULL }
#define	PCCARD_PRODUCT_YEDATA_EXTERNAL_FDD	-1
#define	PCCARD_STR_YEDATA_EXTERNAL_FDD	"Y-E DATA External FDD"
#define	PCCARD_CIS_DIGITAL_DEPCMXX	{ "DIGITAL", "DEPCM-XX", NULL, NULL }
#define	PCCARD_PRODUCT_DIGITAL_DEPCMXX	-1
#define	PCCARD_STR_DIGITAL_DEPCMXX	"DEC DEPCM-BA"
#define	PCCARD_CIS_TEAC_IDECARDII	{ NULL, "NinjaATA-", NULL, NULL }
#define	PCCARD_PRODUCT_TEAC_IDECARDII	-1
#define	PCCARD_STR_TEAC_IDECARDII	"TEAC IDE Card/II"
#define	PCCARD_CIS_LINKSYS_ECARD_2	{ "LINKSYS", "E-CARD", NULL, NULL }
#define	PCCARD_PRODUCT_LINKSYS_ECARD_2	-1
#define	PCCARD_STR_LINKSYS_ECARD_2	"Linksys E-Card"
#define	PCCARD_CIS_COREGA_ETHER_PCC_T	{ "corega K.K.", "corega Ether PCC-T", NULL, NULL }
#define	PCCARD_PRODUCT_COREGA_ETHER_PCC_T	-1
#define	PCCARD_STR_COREGA_ETHER_PCC_T	"Corega"
#define	PCCARD_CIS_COREGA_ETHER_II_PCC_T	{ "corega K.K.", "corega EtherII PCC-T", NULL, NULL }
#define	PCCARD_PRODUCT_COREGA_ETHER_II_PCC_T	-1
#define	PCCARD_STR_COREGA_ETHER_II_PCC_T	"Corega"
#define	PCCARD_CIS_COREGA_FAST_ETHER_PCC_TX	{ "corega K.K.", "corega FastEther PCC-TX", NULL, NULL }
#define	PCCARD_PRODUCT_COREGA_FAST_ETHER_PCC_TX	-1
#define	PCCARD_STR_COREGA_FAST_ETHER_PCC_TX	"Corega"
#define	PCCARD_CIS_SVEC_COMBOCARD	{ "Ethernet", "Adapter", NULL, NULL }
#define	PCCARD_PRODUCT_SVEC_COMBOCARD	-1
#define	PCCARD_STR_SVEC_COMBOCARD	"SVEC/Hawking Tech. Combo Card"
#define	PCCARD_CIS_SVEC_LANCARD	{ "SVEC", "FD605 PCMCIA EtherNet Card", "V1-1", NULL }
#define	PCCARD_PRODUCT_SVEC_LANCARD	-1
#define	PCCARD_STR_SVEC_LANCARD	"SVEC PCMCIA Lan Card"

#define	PCCARD_CIS_NAKAGAWAMETAL_LNT10TN	{ "PCMCIA", "LNT-10TN", NULL, NULL }
#define	PCCARD_PRODUCT_NAKAGAWAMETAL_LNT10TN	-1
#define	PCCARD_STR_NAKAGAWAMETAL_LNT10TN	"NAKAGAWA METAL LNT-10TN NE2000 Compatible Card"

#define	PCCARD_CIS_AMBICOM_AMB8002T	{ "AmbiCom Inc", "AMB8002T", NULL, NULL }
#define	PCCARD_PRODUCT_AMBICOM_AMB8002T	-1
#define	PCCARD_STR_AMBICOM_AMB8002T	"AmbiCom AMB8002T"
#define	PCCARD_CIS_IODATA_PCLAT	{ "I-O DATA", "PCLA", "ETHERNET", NULL }
#define	PCCARD_PRODUCT_IODATA_PCLAT	-1
#define	PCCARD_STR_IODATA_PCLAT	"IO-DATA PCLA/T"
