\ Loader.rc
\ $FreeBSD: src/sys/boot/forth/loader.rc,v 1.2 1999/11/24 17:59:37 dcs Exp $
\
\ Includes additional commands
include /boot/loader.4th

\ Reads and processes loader.rc
start

\ Tests for password -- executes autoboot first if a password was defined
check-password

\ Unless set otherwise, autoboot is automatic at this point

